'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var oscriptAstWalker = require('oscript-ast-walker');
var fs = require('fs');

class Scope {
  constructor (upper, init) {
    this._upper = upper;
    this._vars = new Map();
    if (init) for (const name in init) this.setOwn(name, init[name]);
  }
  get upper () { return this._upper }
  hasOwn (name) {
    return this._vars.has(name)
  }
  has (name) {
    return !!this._findScope(name)
  }
  getOwn (name) {
    return this._vars.get(name)
  }
  get (name) {
    const scope = this._findScope(name);
    return scope && scope.getOwn(name)
  }
  setOwn (name, value) {
    this._vars.set(name, value);
  }
  set (name, value) {
    const scope = this._findScope(name);
    if (scope) scope.setOwn(name, value);
  }
  _findScope (name) {
    let current = this;
    do {
      if (current.hasOwn(name)) return current
      current = current._upper;
    } while (current)
  }
}

const push = Array.prototype.push;
function ignore () {}
const interpreter = {};
interpreter.Program = (node, scope, walk) => walk(node.body, node, scope);
interpreter.PackageDeclaration = () => {
  throw new Error('package is not implemented')
};
interpreter.ObjectName = ({ raw }, scope) => scope.get(raw);
interpreter.ScriptSource = (node, scope, walk) => {
  for (const routine of node.routines) walk(routine, node, scope);
  for (const variable of node.variables) walk(variable, node, scope);
  for (const statement of node.statements) walk(statement, node, scope);
};
interpreter.DumpSource = () => {
  throw new Error('dump is not implemented')
};
interpreter.ScriptDeclaration = interpreter.FunctionDeclaration = (node, scope) => {
  scope.setOwn(node.id.value, node);
};
interpreter.IfStatement = (node, scope, walk) => {
  if (walk(node.test, node, scope)) {
    for (const statement of node.consequent) walk(statement, node, scope);
  } else {
    if (tryElseIfClauses(node.otherClauses, scope, walk)) return
    for (const statement of node.alternate) walk(statement, node, scope);
  }
};
function tryElseIfClauses (otherClauses, scope, walk) {
  for (const otherClause of otherClauses) {
    if (walk(otherClause.test, otherClause, scope)) {
      for (const statement of otherClause.consequent) walk(statement, otherClause, scope);
      return true
    }
  }
}
interpreter.SwitchStatement = (node, scope, walk) => {
  const discriminant = walk(node.discriminant, node, scope);
  node.cases.some(switchCase => trySwitchCase(discriminant, switchCase, scope, walk));
};
function trySwitchCase (discriminant, node, scope, walk) {
  if (!node.tests || node.tests.some(test => walk(test, node, scope) === discriminant)) {
    for (const statement of node.consequent) walk(statement, node, scope);
    return true
  }
}
interpreter.WhileStatement = (node, scope, walk) => {
  while (walk(node.test, node, scope)) {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  }
};
interpreter.RepeatStatement = (node, scope, walk) => {
  do {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  } while (walk(node.test, node, scope))
};
interpreter.ForStatement = (node, scope, walk) => {
  for (node.init && walk(node.init, node, scope);
    !node.test || walk(node.test, node, scope);
    node.update && walk(node.update, node, scope)) {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  }
};
interpreter.ForEachStatement = (node, scope, walk) => {
  const name = node.left.value;
  for (const item of walk(node.right, node, scope)) {
    scope.set(name, item);
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
  }
};
interpreter.StructuredForStatement = (node, scope, walk) => {
  const name = node.variable.value;
  let index = walk(node.start, node, scope);
  const end = walk(node.end, node, scope);
  const step = node.step ? walk(node.step, node, scope) : 1;
  for (scope.set(name, index); index < end;) {
    scope.break = scope.continue = false;
    for (const statement of node.body) {
      walk(statement, node, scope);
      if (scope.break) return
      if (scope.continue) break
    }
    scope.set(name, node.down ? index -= step : index += step);
  }
};
interpreter.BreakStatement = (node, scope) => { scope.break = true; };
interpreter.ContinueStatement = (node, scope) => { scope.continue = true; };
interpreter.BreakIfStatement = (node, scope, walk) => {
  scope.break = walk(node.test, node, scope);
};
interpreter.ContinueIfStatement = (node, scope, walk) => {
  scope.continue = walk(node.test, node, scope);
};
interpreter.GotoStatement = (node, scope, walk) => {
  throw new Error('package is not implemented')
};
interpreter.LabelStatement = interpreter.EmptyStatement = ignore;
interpreter.ReturnStatement = (node, scope, walk) => {
  if (node.argument) scope.result = walk(node.argument, node, scope);
};
interpreter.VariableDeclaration = (node, scope, walk) => {
  for (const declaration of node.declarations) {
    const { init } = declaration;
    scope.setOwn(declaration.id.value, init && walk(init, declaration, scope));
  }
};
interpreter.ConditionalExpression = (node, scope, walk) => {
  return walk(node.test, node, scope)
    ? walk(node.consequent, node, scope)
    : walk(node.alternate, node, scope)
};
interpreter.BinaryExpression = (node, scope, walk) => {
  let { left, operator, right } = node;
  left = isAssignment(operator) ? getName(left) : walk(left, node, scope);
  return binaryOperation[operator](left, walk(right, node, scope))
};
function isAssignment (symbol) {
  switch (symbol.length) {
    case 1:
      return symbol === '='
    case 2:
      return symbol === '+=' || symbol === '-=' || symbol === '*=' ||
        symbol === '&=' || symbol === '|=' || symbol === '^='
  }
  return false
}
function getName (node) {
  const { type } = node;
  return type === 'Identifier'
    ? node.value
    : type === 'ObjectName' || type === 'LegacyAlias'
      ? node.raw
      : 'dynamic'
}
const binaryOperation = {
  '=': (left, right, scope) => scope.set(left, right),
  '+': (left, right) => left + right,
  '-': (left, right) => left - right,
  '*': (left, right) => left * right,
  '/': (left, right) => left / right,
  '<': (left, right) => left < right,
  '>': (left, right) => left > right,
  '%': (left, right) => left % right,
  '&': (left, right) => left & right,
  '|': (left, right) => left | right,
  '^': (left, right) => left ^ right,
  '==': (left, right) => left === right,
  '!=': (left, right) => left !== right,
  '<>': (left, right) => left !== right,
  '<=': (left, right) => left <= right,
  '>=': (left, right) => left >= right,
  '&&': (left, right) => left && right,
  in: (left, right) => left in right,
  '+=': (left, right, scope) => scope.set(left, scope.get(left) + right),
  '-=': (left, right, scope) => scope.set(left, scope.get(left) - right),
  '*=': (left, right, scope) => scope.set(left, scope.get(left) * right),
  '&=': (left, right, scope) => scope.set(left, scope.get(left) & right),
  '|=': (left, right, scope) => scope.set(left, scope.get(left) | right),
  '^=': (left, right, scope) => scope.set(left, scope.get(left) ^ right),
  '^^': (left, right) => !!(left ^ right),
  '<<': (left, right) => left << right,
  '>>': (left, right) => left >> right,
  or: (left, right) => left || right,
  eq: (left, right) => left === right,
  ne: (left, right) => left !== right,
  lt: (left, right) => left < right,
  le: (left, right) => left <= right,
  gt: (left, right) => left > right,
  ge: (left, right) => left >= right,
  and: (left, right) => left && right,
  xor: (left, right) => !!(left ^ right)
};
interpreter.UnaryExpression = (node, scope, walk) =>
  unaryOperation[node.operator](walk(node.argument, node, scope));
const unaryOperation = {
  '!': value => !value,
  '+': value => value,
  '-': value => -value,
  '~': value => ~value,
  not: value => !value
};
interpreter.MemberExpression = (node, scope, walk) =>
  walk(node.object, node, scope)[walk(node.property, node, scope)];
interpreter.SliceExpression = (node, scope, walk) => {
  const array = walk(node.object, node, scope);
  const start = (node.start && walk(node.start, node, scope)) || 0;
  const end = (node.end && walk(node.end, node, scope)) || array.length;
  return array.slice(start, end)
};
interpreter.IndexExpression = (node, scope, walk) =>
  walk(node.object, node, scope)[walk(node.index, node, scope)];
interpreter.CallExpression = (node, scope, walk) => {
  const routine = walk(node.callee, node, scope);
  if (!routine) throw new Error(`routine "${getName(node.callee)}" is undefined`)
  const values = node.arguments.map(arg => walk(arg, node, scope));
  if (typeof routine === 'function') return routine(...values)
  if (!routine.type) throw new Error(`"${getName(node.callee)}" is not a routine`)
  const args = routine.params.reduce((args, param, index) => {
    const value = values[index];
    args[param.id.value] = value === undefined && param.init
      ? walk(param.init, node, scope)
      : value;
    return args
  }, {});
  scope = new Scope(scope, args);
  scope.result = undefined;
  if (node.routines) for (const routine of node.routines) walk(routine, node, scope);
  for (const variable of node.variables) walk(variable, node, scope);
  for (const statement of node.statements) walk(statement, node, scope);
  return scope.result
};
interpreter.ThisExpression = (node, scope) => scope;
interpreter.SuperExpression = (node, { upper }) => upper;
interpreter.AssocExpression = interpreter.ObjectExpression = (node, scope, walk) => {
  return node.properties.reduce((object, { key, value }) => {
    object[walk(key, node, scope)] = walk(value, node, scope);
    return object
  }, {})
};
interpreter.ListExpression = (node, scope, walk) =>
  node.elements.reduce((result, element) => {
    const item = walk(element, node, scope);
    if (element.type === 'AtExpression') push.apply(result, item);
    else result.push(item);
    return result
  }, []);
interpreter.ListComprehension = (node, scope, walk) => {
  const name = node.left.value;
  const values = walk(node.right, node, scope);
  return values.reduce((result, value) => {
    scope.set(name, value);
    if (!node.test || walk(node.test, node, scope)) {
      result.push(walk(node.expression, node, scope));
    }
    return result
  }, [])
};
interpreter.AtExpression = interpreter.ParenthesisExpression =
  (node, scope, walk) => walk(node.expression, node, scope);
interpreter.XlateExpression = ({ raw }, scope) => scope.get(raw);
interpreter.Identifier = ({ value }, scope) => scope.get(value);
interpreter.LegacyAlias = ({ raw }, scope) => scope.get(raw);
interpreter.Literal = ({ value }) => value;

function optimize (ast) {
  oscriptAstWalker.simple(ast, optimizer);
  return ast
}
const optimizer = {};
optimizer.ScriptSource = optimizer.ScriptDeclaration = {
  pre (node) {
    const routines = node.routines = [];
    const variables = node.variables = [];
    const statements = node.statements = [];
    for (const part of node.body) {
      const { type } = part;
      const target = type === 'VariableDeclaration'
        ? variables
        : type === 'FunctionDeclaration' || type === 'ScriptDeclaration'
          ? routines
          : statements;
      target.push(part);
    }
  }
};
optimizer.FunctionDeclaration = {
  pre (node) {
    const variables = node.variables = [];
    const statements = node.statements = [];
    for (const part of node.body) {
      const { type } = part;
      const target = type === 'VariableDeclaration' ? variables : statements;
      target.push(part);
    }
  }
};
optimizer.ObjectName = {
  pre (node) {
    node.raw = node.name
      .map(({ value }) => typeof value === 'number' ? `&${value.toString(16)}` : value)
      .join('::');
    return true
  }
};
optimizer.LegacyAlias = {
  pre (node) {
    node.raw = `&${node.value.toString(16)}`;
    return true
  }
};
optimizer.XlateExpression = {
  pre (node) {
    node.raw = `${node.ospace.value}.${node.string.value}`;
    return true
  }
};

const cr = '\r';
const crlf = '\r\n';
const lf = '\n';
const tab = '\t';
function ascii (value) {
  return typeof value === 'number' ? String.fromCharCode(value) : value.charCodeAt(0)
}
function bytelength (text) {
  return typeof Buffer !== 'undefined'
    ? Buffer.from(text).length
    : new TextEncoder().encode(text).length
}
function capitalize (text) {
  return text.replace(/(\w+)/g, match =>
    match.charAt(0).toUpperCase() + match.substr(1).toLowerCase())
}
function catenate (values, delimiter) {
  const suffix = values.length ? delimiter : '';
  return `${values.join(delimiter)}${suffix}`
}
function chr (hay, needle) {
  const index = hay.indexOf(needle.charAt(0));
  return index < 0 ? undefined : index + 1
}
function cmp (left, right, length) {
  if (length !== undefined) {
    left = left.substr(0, length);
    right = right.substr(0, length);
  }
  return left < right ? -1 : left > right ? 1 : 0
}
function cmpbe (left, right, length) {
  return cmp(left.trimEnd(), right.trimEnd(), length)
}
function cmpi (left, right, length) {
  return cmp(left.toLowerCase(), right.toLowerCase(), length)
}
function cmpibe (left, right, length) {
  return cmp(left.trimEnd().toLowerCase(), right.trimEnd().toLowerCase(), length)
}
function collapse (text) {
  return text.replace(/[ \t]/g, '')
}
function compress (text) {
  return text.replace(/[ \t]+/g, ' ')
}
function cspn (text, chars) {
  const regexp = new RegExp(`^[^${chars.split('').map(escapeRegExp).join('|')}]+`);
  const match = regexp.exec(text);
  return match ? match[0].length : 0
}
function cstring (text, quote = '\'') {
  return text.replace(new RegExp(`[\r\n\t\\${quote}]`, 'g'), '\\$1')
}
function elements (text, delimiter) {
  return text.split(delimiter.charAt(0))
}
function filetostring (name) {
  return fs.readFileSync(name, 'utf8')
}
function format (format, ...args) {
  return format.replace(/%(\d)/g, (match, index) => '' + args[index - 1])
}
function frombase64 (text) {
  return typeof Buffer !== 'undefined'
    ? Buffer.from(text, 'base64').toString('ascii')
    : atob(text)
}
function hyphenate (text, length) {
  const parts = [];
  do {
    parts.push(text.substr(0, length));
    text = text.substr(length);
  } while (text.length)
  return parts
}
function join (values, delimiter) {
  return values.join(delimiter)
}
function locate (hay, needle) {
  const index = hay.indexOf(needle);
  return index < 0 ? undefined : index + 1
}
function locatei (hay, needle) {
  const index = hay.toLowerCase().indexOf(needle.toLowerCase());
  return index < 0 ? undefined : index + 1
}
function lower (text) {
  return text.toLowerCase()
}
function quote (text, quote = '\'') {
  return `${quote}${text.replace(new RegExp(quote, 'g'), `\\${quote}`)}${quote}`
}
function rchr (hay, needle) {
  const index = hay.lastIndexOf(needle.charAt(0));
  return index < 0 ? undefined : index + 1
}
function replace (text, find, replace) {
  return text.replace(find, replace)
}
function replaceall (text, find, replace) {
  const regexp = new RegExp(`${find.split('').map(escapeRegExp).join('')}`, 'g');
  return text.replace(regexp, replace)
}
function set (length, text) {
  return text.charAt(0).repeat(length)
}
function spn (text, chars) {
  const regexp = new RegExp(`^[${chars.split('').map(escapeRegExp).join('')}]+`);
  const match = regexp.exec(text);
  return match ? match[0].length : 0
}
function string (value, raw) {
  return !raw && value instanceof Error ? value.message : valuetostring(value)
}
function stringtointegher (text) {
  return parseInt(text)
}
function stringtoreal (text) {
  return parseFloat(text)
}
function stringtovalue (text) {
  const number = parseInt(text);
  if (isNaN(number)) return number
  const normalized = text.toLowerCase();
  if (normalized === 'true') return true
  if (normalized === 'false') return false
  if (normalized === 'undefined') return undefined
  return text
}
function strip (text, chars) {
  const regexp = new RegExp(`[${chars.split('').map(escapeRegExp).join('')}]`, 'g');
  return text.replace(regexp, '')
}
function tobase64 (text) {
  return typeof Buffer !== 'undefined'
    ? Buffer.from(text).toString('base64')
    : btoa(text)
}
function upper (text) {
  return text.toUpperCase()
}
function valuetostring (value) {
  return String(value)
}
function escapeRegExp (char) {
  return char === '*' || char === '+' || char === '.' || char === '?' ||
      char === '[' || char === ']' || char === '(' || char === ')' ||
      char === '{' || char === '}' || char === '^' || char === '$'
    ? `\\${char}`
    : char
}

var str = /*#__PURE__*/Object.freeze({
  __proto__: null,
  cr: cr,
  crlf: crlf,
  lf: lf,
  tab: tab,
  ascii: ascii,
  bytelength: bytelength,
  capitalize: capitalize,
  catenate: catenate,
  chr: chr,
  cmp: cmp,
  cmpbe: cmpbe,
  cmpi: cmpi,
  cmpibe: cmpibe,
  collapse: collapse,
  compress: compress,
  cspn: cspn,
  cstring: cstring,
  elements: elements,
  filetostring: filetostring,
  format: format,
  frombase64: frombase64,
  hyphenate: hyphenate,
  join: join,
  locate: locate,
  locatei: locatei,
  lower: lower,
  quote: quote,
  rchr: rchr,
  replace: replace,
  replaceall: replaceall,
  set: set,
  spn: spn,
  string: string,
  stringtointegher: stringtointegher,
  stringtoreal: stringtoreal,
  stringtovalue: stringtovalue,
  strip: strip,
  tobase64: tobase64,
  upper: upper,
  valuetostring: valuetostring
});

const crlf$1 = '\r\n';
const js_encodeuri = '\n';
const js_encodeuricomponent = '\t';
const js_escape = '\t';
function decodeforurl (text) {
  return decodeURI(text)
}
function encodeforurl (text) {
  return encodeURI(text)
}
function escape (text) {
  return encodeURIComponent(text)
}
function escapeforjs (text, mode) {
  return mode === js_escape
    ? escape(text)
    : mode === js_encodeuri
      ? encodeURI(text)
      : encodeURIComponent(text)
}
function escapehtml (text) {
  return text
    .replace(/&/g, '&amp;')
    .replace(/</g, '&lt;')
    .replace(/</g, '&gt;')
}
function escapejson (text, trim) {
  if (trim) text = text.trim();
  const string = JSON.stringify(text);
  return string.substr(1, string.length - 2)
}
function escapexml (text) {
  return text
    .replace(/&/g, '&amp;')
    .replace(/</g, '&lt;')
    .replace(/</g, '&gt;')
    .replace(/'/g, '&apos;')
}
function tojson (value, trim = true) {
  return JSON.parse(trimStrings(value))
}
function unescape (text) {
  return decodeURIComponent(text)
}
function unescapejson (text) {
  return JSON.parse(`"${text}}"`)
}
function trimStrings (value) {
  if (typeof value === 'string') return value.trim()
  if (Array.isArray(value)) {
    for (let i = 0, l = value.length; i < l; ++i) value[i] = trimStrings(value[i]);
  }
  if (value && typeof value === 'object') {
    for (const key in value) value[key] = trimStrings(value[key]);
  }
  return value
}

var web = /*#__PURE__*/Object.freeze({
  __proto__: null,
  crlf: crlf$1,
  js_encodeuri: js_encodeuri,
  js_encodeuricomponent: js_encodeuricomponent,
  js_escape: js_escape,
  decodeforurl: decodeforurl,
  encodeforurl: encodeforurl,
  escape: escape,
  escapeforjs: escapeforjs,
  escapehtml: escapehtml,
  escapejson: escapejson,
  escapexml: escapexml,
  tojson: tojson,
  unescape: unescape,
  unescapejson: unescapejson
});

const { log, debug, error, info, warn } = console;
let stamp;
function pad2 (n) { return n < 10 ? `0${n}` : n }
function pad3 (n) { return n < 10 ? `00${n}` : n < 100 ? `0${n}` : n }
function formatStamp () {
  const date = new Date();
  return pad2((date.getMonth() + 1)) + '/' + pad2(date.getDate()) + '/' +
    (date.getYear() + 1900) + ' ' + pad2(date.getHours()) + ':' +
    pad2(date.getMinutes()) + ':' + pad2(date.gitSeconds()) + '.' +
    pad3(date.getMilliseconds())
}
function formatMessage (...args) {
  const message = args.join('');
  return stamp ? `${formatStamp()} ${message}` : message
}
function echo (...args) {
  log(formatMessage(...args));
}
function echodebug (...args) {
  debug(formatMessage(...args));
}
function echoerror (...args) {
  error(formatMessage(...args));
}
function echoinfo (...args) {
  info(formatMessage(...args));
}
function echostamp (enable) {
  stamp = enable;
}
function echowarn (...args) {
  warn(formatMessage(...args));
}
function isdefined (value) {
  return value !== undefined
}
function iserror (value) {
  return value instanceof Error
}
function isfeature (object, property) {
  return property in object
}
function isinvokable (value) {
  return typeof value === 'function' || !!value.type
}
function isnoterror (value) {
  return !(value instanceof Error)
}
function isundefined (value) {
  return value === undefined
}
function length (value) {
  return value && value.length
}

var library = /*#__PURE__*/Object.freeze({
  __proto__: null,
  str: str,
  web: web,
  echo: echo,
  echodebug: echodebug,
  echoerror: echoerror,
  echoinfo: echoinfo,
  echostamp: echostamp,
  echowarn: echowarn,
  isdefined: isdefined,
  iserror: iserror,
  isfeature: isfeature,
  isinvokable: isinvokable,
  isnoterror: isnoterror,
  isundefined: isundefined,
  length: length
});

function interpret (ast, globals = {}) {
  const scope = new Scope(null, Object.assign({}, library, globals));
  oscriptAstWalker.recursive(optimize(ast), null, interpreter, scope);
}

exports.interpret = interpret;
//# sourceMappingURL=index.js.map
